__version_info__ = {
    "major": 1,
    "minor": 2,
    "micro": 0,
    "releaselevel": "final",
    "serial": 4,
}


def get_version(short=False):
    assert __version_info__["releaselevel"] in ("alpha", "beta", "final")
    vers = ["%(major)i.%(minor)i" % __version_info__]
    if __version_info__["micro"]:
        vers.append(".%(micro)i" % __version_info__)
    if __version_info__["releaselevel"] != "final" and not short:
        vers.append("%s%i" % (__version_info__["releaselevel"][0], __version_info__["serial"]))
    return "".join(vers)


__version__ = get_version()



from uuid import uuid4

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from .utils import get_email_from_token, get_login_url, get_logout_url, is_email_verified_from_token

try:
    from django.contrib.auth import get_user_model
except ImportError:
    from django.contrib.auth.models import User

    def get_user_model(*args, **kwargs):
        return User


if getattr(settings, "SESSION_COOKIE_SAMESITE"):
    raise ImproperlyConfigured(
        """Please set the SESSION_COOKIE_SAMESITE to None in your settings.py. To prevent replay 
        attacks and forgery the callback URL verifies a unique state string and nonce string stored 
        the session is equal value that is posted back by Okta."""
    )


class OktaBackend(object):
    USER_CREATION = getattr(settings, "OKTA_USER_CREATION", True)

    supports_anonymous_user = False
    supports_inactive_user = True
    supports_object_permissions = False

    def __init__(self):
        self.User = get_user_model()

    def login_url(self, redirect_uri, state, nonce):
        return get_login_url(redirect_uri=redirect_uri, state=state, nonce=nonce)

    def logout_url(self, id_token, redirect_uri):
        return get_logout_url(id_token=id_token, redirect_uri=redirect_uri)

    def authenticate(self, token=None, **kwargs):
        if token is None:
            return None

        email = get_email_from_token(token=token)

        if email is None:
            return None

        email_verified = is_email_verified_from_token(token=token)

        if not email_verified:
            return None

        users = self.User.objects.filter(email=email)
        if len(users) == 0:
            user = self.create_user(email)
            if user is None:
                return None
        elif len(users) == 1:
            user = users[0]
        else:
            return None
        if not self.user_can_authenticate(user):
            return None
        user.backend = "{}.{}".format(self.__class__.__module__, self.__class__.__name__)
        return user

    def user_can_authenticate(self, user):
        is_active = getattr(user, "is_active", None)
        return is_active or is_active is None

    def get_user(self, user_id):
        try:
            user = self.User.objects.get(pk=user_id)
            return user
        except self.User.DoesNotExist:
            return None

    def create_user(self, email):
        if self.USER_CREATION:
            username_field = getattr(self.User, "USERNAME_FIELD", "username")
            user_kwargs = {"email": email}
            user_kwargs[username_field] = self.username_generator(email)
            return self.User.objects.create_user(**user_kwargs)
        else:
            return None

    @staticmethod
    def username_generator(email):
        return str(uuid4())



from django.urls import re_path

from .views import auth, callback, logout

app_name = "okta"
urlpatterns = [
    re_path(r"^login/$", auth, name="login"),
    re_path(r"^logout/$", logout, name="logout"),
    re_path(r"^callback/$", callback, name="callback"),
]



import logging

import jwt
from cryptography.hazmat.primitives import serialization
from django.conf import settings
from jwt import PyJWKClient

from .cache import lru_cache_with_expiration

try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode


logger = logging.getLogger("okta_auth")


DOMAIN = getattr(settings, "OKTA_DOMAIN")
SCOPE = getattr(settings, "OKTA_SCOPE", "openid email")
RESPONSE_TYPE = getattr(settings, "OKTA_RESPONSE_TYPE", "id_token")
CLIENT_ID = getattr(settings, "OKTA_CLIENT_ID")
CLIENT_SECRET = getattr(settings, "OKTA_CLIENT_SECRET")


def get_login_url(
    domain=DOMAIN,
    scope=SCOPE,
    client_id=CLIENT_ID,
    redirect_uri=None,
    response_type=RESPONSE_TYPE,
    response_mode="form_post",
    state=None,
    nonce=None,
):
    param_dict = {
        "response_type": response_type,
        "response_mode": response_mode,
        "scope": scope,
        "nonce": "nonce",
        "client_id": client_id,
    }
    if redirect_uri is not None:
        param_dict["redirect_uri"] = redirect_uri
    if state is not None:
        param_dict["state"] = state
    if nonce is not None:
        param_dict["nonce"] = nonce
    params = urlencode(param_dict)
    return "https://{domain}/oauth2/v1/authorize?{params}".format(domain=domain, params=params)


def get_logout_url(id_token, redirect_uri, domain=DOMAIN):
    params = urlencode({"post_logout_redirect_uri": redirect_uri, "id_token_hint": id_token})
    return "https://{domain}/oauth2/v1/logout?{params}".format(domain=domain, params=params)


@lru_cache_with_expiration(seconds=60 * 60 * 24)
def _get_payload(token, domain=DOMAIN, audience=CLIENT_ID):
    url = "https://{domain}/oauth2/v1/keys".format(domain=domain)
    jwks_client = PyJWKClient(url)
    signing_key = jwks_client.get_signing_key_from_jwt(token)
    pem_key = signing_key.key.public_bytes(
        encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    try:
        payload = jwt.decode(
            token, key=pem_key, algorithms=["RS256"], audience=audience, leeway=300
        )
        return payload
    except jwt.InvalidTokenError as e:
        logger.debug("Could not retrieve payload. Token validation error, {}".format(str(e)))

    return {}


def get_email_from_token(token=None, audience=CLIENT_ID):
    try:
        payload = _get_payload(token=token, audience=audience)
        if "email" in payload:
            return payload["email"]
        elif "sub" in payload:
            return payload["sub"].split("|").pop()
        else:
            logger.debug(
                'Could not retrieve email. Token payload does not contain keys: "email" or "sub".'
            )
    except jwt.InvalidTokenError as e:
        logger.debug("Could not retrieve email. Token validation error, {}".format(str(e)))

    return None


def is_email_verified_from_token(token=None, audience=CLIENT_ID):
    try:
        payload = _get_payload(token=token, audience=audience)
        return payload.get("email_verified", True)
    except jwt.InvalidTokenError as e:
        logger.debug(
            "Could not determine email verification status. Token validation error, {}".format(
                str(e)
            )
        )

    return None


def get_nonce_from_token(token=None, audience=CLIENT_ID):
    try:
        payload = _get_payload(token=token, audience=audience)
        return payload.get("nonce")
    except jwt.InvalidTokenError as e:
        logger.debug("Could not retrieve nonce. Token validation error, {}".format(str(e)))

    return None



import os

try:
    from setuptools import find_packages, setup
except ImportError:
    from distutils.core import find_packages, setup


def read_file(filename):
    """Read a file into a string"""
    path = os.path.abspath(os.path.dirname(__file__))
    filepath = os.path.join(path, filename)
    try:
        return open(filepath).read()
    except IOError:
        return ""


def get_readme():
    """Return the README file contents. Supports text,rst, and markdown"""
    for name in ("README", "README.rst", "README.md"):
        if os.path.exists(name):
            return read_file(name)
    return ""


setup(
    name="django-okta-authentication",
    version=__import__("okta_auth").get_version().replace(" ", "-"),
    url="https://github.com/jasonchrista/django-okta-auth",
    author="Jason Christa",
    author_email="jason@zeitcode.com",
    description="Authenticated users using Okta",
    long_description=get_readme(),
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    install_requires=read_file("requirements.txt"),
    classifiers=[
        "Environment :: Web Environment",
        "License :: OSI Approved :: BSD License",
        "Framework :: Django",
        "Programming Language :: Python",
    ],
)



from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME, login
from django.contrib.auth import logout as auth_logout
from django.core.exceptions import PermissionDenied
from django.http import HttpResponseRedirect
from django.shortcuts import redirect, resolve_url

from .backends import OktaBackend
from .utils import get_nonce_from_token

try:
    from django.core.urlresolvers import reverse
except ImportError:
    from django.urls import reverse

import logging

from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_exempt

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

import uuid

logger = logging.getLogger("okta_auth")


@never_cache
def auth(request):
    backend = OktaBackend()
    redirect_uri = request.build_absolute_uri(reverse("okta:callback"))
    redirect_to = request.GET.get(REDIRECT_FIELD_NAME, "")
    if redirect_to:
        # redirect_uri = "{}?{}={}".format(redirect_uri, REDIRECT_FIELD_NAME, redirect_to)
        request.session[REDIRECT_FIELD_NAME] = redirect_to
    state = str(uuid.uuid4())
    request.session["state"] = state
    nonce = str(uuid.uuid4())
    request.session["nonce"] = nonce
    login_url = backend.login_url(
        redirect_uri=redirect_uri,
        state=state,
        nonce=nonce,
    )
    return HttpResponseRedirect(login_url)


@never_cache
def logout(request):
    backend = OktaBackend()
    id_token = request.session.get("id_token")
    logout_redirect_url = getattr(settings, "LOGOUT_REDIRECT_URL", "/")
    redirect_uri = request.build_absolute_uri(resolve_url(logout_redirect_url))
    logout_url = backend.logout_url(
        id_token=id_token,
        redirect_uri=redirect_uri,
    )
    auth_logout(request)
    return HttpResponseRedirect(logout_url)


@never_cache
@csrf_exempt
def callback(request):
    login_response = redirect("okta:login")
    backend = OktaBackend()

    original_state = request.session.get("state")
    state = request.POST.get("state")
    if original_state != state:
        logger.debug("Expected state {} but received {}.".format(original_state, state))
        return login_response

    token = request.POST.get("id_token")
    request.session["id_token"] = token
    logger.debug("Token {} received".format(token))

    original_nonce = request.session.get("nonce")
    nonce = get_nonce_from_token(token=token)
    if original_nonce != nonce:
        logger.debug("Expected nonce {} but received {}.".format(original_nonce, nonce))
        return login_response

    user = backend.authenticate(token=token)
    if user is not None:
        login(request, user)
        return HttpResponseRedirect(get_login_success_url(request))
    else:
        logger.debug("Authenticated user not in user database.")
        raise PermissionDenied()


def get_login_success_url(request):
    # redirect_to = request.GET.get(REDIRECT_FIELD_NAME, "")
    redirect_to = request.session.pop(REDIRECT_FIELD_NAME, "")
    netloc = urlparse(redirect_to)[1]
    if not redirect_to:
        redirect_to = settings.LOGIN_REDIRECT_URL
    elif netloc and netloc != request.get_host():
        redirect_to = settings.LOGIN_REDIRECT_URL
    return redirect_to
